package net.zenrindatacom.geofenceapp_sample.http;

import android.util.Base64;
import android.util.Log;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import net.zenrindatacom.geofenceapp_sample.http.dto.SrvSearchAreaReqDto;
import net.zenrindatacom.geofenceapp_sample.http.dto.SrvSearchAreaResDto;
import net.zenrindatacom.geofenceapp_sample.http.dto.SrvSearchAreaUpdateTimeResDto;
import net.zenrindatacom.geofenceapp_sample.http.dto.SrvTokenReqDto;
import net.zenrindatacom.geofenceapp_sample.http.dto.SrvTokenResDto;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.Map;

import javax.net.ssl.HttpsURLConnection;

/**
 * サーバーアクセスクラス
 *
 * @author N.Miyahara
 */
public class ServerAccess {

    private static class ApiResult {
        /** HTTPステータスコード */
        private int httpStatusCode = 0;
        /** レスポンスヘッダ */
        @SuppressWarnings("MismatchedQueryAndUpdateOfCollection")
        private final Map<String, String> resHeader = new HashMap<>();
        /** レスポンスボディ */
        private String rawResBody = null;
    }

    /**
     * HTTPメソッド
     */
    @SuppressWarnings("unused")
    private enum Method {
        GET,
        POST,
        PUT,
        DELETE
    }

    /** タグ */
    private static final String TAG = "ServerAccess";

    /** コネクションタイムアウト時間(ミリ秒) */
    private static final int CONNECTION_TIMEOUT = 60 * 1000;

    /** リードタイムアウト時間(ミリ秒) */
    private static final int READ_TIMEOUT = 60 * 1000;

    /** インスタンス */
    private static final ServerAccess self = new ServerAccess();

    /**
     * コンストラクタ
     */
    private ServerAccess() {}

    /**
     * シングルトンインスタンス取得
     */
    public static ServerAccess getInstance() {
        return self;
    }

    /**
     * (トークン発行サーバ)トークン取得API
     *
     * @param url リクエスト先のURL
     * @param clientId クライアントID
     * @param secretKey 秘密鍵
     * @param srvTokenReqDto リクエストデータ
     * @return トークン取得APIの取得結果(DTO)
     * @throws TokenServerException トークン発行サーバ例外
     */
    public SrvTokenResDto getToken(final String url,
                                   final String clientId,
                                   final String secretKey,
                                   final SrvTokenReqDto srvTokenReqDto) throws TokenServerException {
        //実行メソッド名
        final String methodName = "getToken";

        //=========== URLの生成 ===========
        URL reqUrl;
        try {
            reqUrl = new URL(url);
        } catch (MalformedURLException e) {
            e.printStackTrace();
            Log.e(TAG, methodName + ": Url Format Invalid.");
            throw new TokenServerException(TokenServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== ヘッダの生成 ===========
        Map<String, String> reqHeader = new HashMap<>();
        reqHeader.put("authorization", "Basic " + ServerAccess.getBase64encode(clientId + ":" + secretKey));
        reqHeader.put("content-type", "application/x-www-form-urlencoded");

        //=========== ボディの生成 ===========
        String reqBody = srvTokenReqDto.getRequestBody();

        //=========== API実行 ===========
        ApiResult apiResult;
        try {
            apiResult = invokeApi(reqUrl, Method.POST, reqHeader, reqBody, HttpsURLConnection.HTTP_OK);
        } catch(SocketTimeoutException e) {
            //タイムアウト
            throw new TokenServerException(TokenServerException.ErrorType.TIMEOUT, e);
        } catch(Exception e) {
            //その他エラー
            throw new TokenServerException(TokenServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== レスポンスの確認 ===========
        SrvTokenResDto srvTokenResDto;

        if (apiResult.httpStatusCode == HttpsURLConnection.HTTP_OK) {
            //正常レスポンス

            if(apiResult.rawResBody == null) {
                //レスポンスボディが空の場合はエラー
                Log.e(TAG,methodName + ":response body is null");
                throw new TokenServerException(TokenServerException.ErrorType.ETC_ERROR);
            }

            try {
                //JSON文字列をDTOにマッピング
                ObjectMapper objectMapper = new ObjectMapper();
                srvTokenResDto = objectMapper.readValue(apiResult.rawResBody, SrvTokenResDto.class);
            } catch (JsonProcessingException e) {
                Log.e(TAG,methodName + ":response body mapping error");
                throw new TokenServerException(TokenServerException.ErrorType.ETC_ERROR, e);
            }

        } else if(apiResult.httpStatusCode == HttpsURLConnection.HTTP_BAD_REQUEST) {
            //サーバーエラー（認証エラー）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new TokenServerException(TokenServerException.ErrorType.SERVER_AUTH_FAILED);

        } else {
            //サーバーエラー（上記以外）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new TokenServerException(TokenServerException.ErrorType.SERVER_ERROR);
        }

        return srvTokenResDto;
    }

    /**
     * (エリア情報サーバ)エリア更新日時検索API
     *
     * @param url リクエスト先のURL
     * @param apiKey APIキー
     * @param tokenType トークンタイプ
     * @param token トークン
     * @return エリア更新日時検索結果(DTO)
     * @throws AreaServerException エリア情報サーバ例外
     */
    public SrvSearchAreaUpdateTimeResDto searchAreaUpdateTime(
            final String url,
            final String apiKey,
            final String tokenType,
            final String token) throws AreaServerException {

        //実行メソッド名
        final String methodName = "searchAreaUpdateTime";

        //=========== URLの生成 ===========
        URL reqUrl;
        try {
            reqUrl = new URL(url);
        } catch (MalformedURLException e) {
            e.printStackTrace();
            Log.e(TAG, methodName + ": Url Format Invalid.");
            throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== ヘッダの生成 ===========
        Map<String, String> reqHeader = new HashMap<>();
        reqHeader.put("x-api-key", apiKey);
        reqHeader.put("authorization", tokenType + " " + token);
        reqHeader.put("content-type", "application/x-www-form-urlencoded");

        //=========== ボディの生成 ===========
        String reqBody = null;

        //=========== API実行 ===========
        ApiResult apiResult;
        try {
            //noinspection ConstantConditions
            apiResult = invokeApi(reqUrl, Method.POST, reqHeader, reqBody, HttpsURLConnection.HTTP_OK);
        } catch(SocketTimeoutException e) {
            //タイムアウト
            throw new AreaServerException(AreaServerException.ErrorType.TIMEOUT, e);
        } catch(Exception e) {
            //その他エラー
            throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== レスポンスの確認 ===========
        SrvSearchAreaUpdateTimeResDto srvSearchAreaUpdateTimeResDto;

        if (apiResult.httpStatusCode == HttpsURLConnection.HTTP_OK) {
            //正常レスポンス

            if(apiResult.rawResBody == null) {
                //レスポンスボディが空の場合はエラー
                Log.e(TAG,methodName + ":response body is null");
                throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR);
            }

            try {
                //JSON文字列をDTOにマッピング
                ObjectMapper objectMapper = new ObjectMapper();
                srvSearchAreaUpdateTimeResDto = objectMapper.readValue(apiResult.rawResBody, SrvSearchAreaUpdateTimeResDto.class);
            } catch (JsonProcessingException e) {
                Log.e(TAG,methodName + ":response body mapping error");
                throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
            }

        } else if(apiResult.httpStatusCode == HttpsURLConnection.HTTP_BAD_REQUEST) {
            //サーバーエラー（パラメータエラー）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new AreaServerException(AreaServerException.ErrorType.SERVER_PARAM_ERROR);

        } else {
            //サーバーエラー（上記以外）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new AreaServerException(AreaServerException.ErrorType.SERVER_ERROR);
        }

        return srvSearchAreaUpdateTimeResDto;
    }

    /**
     * (エリア情報サーバ)エリア検索API
     *
     * @param url リクエスト先のURL
     * @param apiKey APIキー
     * @param tokenType トークンタイプ
     * @param token トークン
     * @param srvSearchAreaReqDto リクエストデータ
     * @return エリア検索結果(DTO)
     * @throws AreaServerException エリアサーバ例外
     */
    public SrvSearchAreaResDto searchArea(
            final String url,
            final String apiKey,
            final String tokenType,
            final String token,
            final SrvSearchAreaReqDto srvSearchAreaReqDto) throws AreaServerException {

        //実行メソッド名
        final String methodName = "searchArea";

        //=========== URLの生成 ===========
        URL reqUrl;
        try {
            reqUrl = new URL(url);
        } catch (MalformedURLException e) {
            e.printStackTrace();
            Log.e(TAG, methodName + ": Url Format Invalid.");
            throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== ヘッダの生成 ===========
        Map<String, String> reqHeader = new HashMap<>();
        reqHeader.put("x-api-key", apiKey);
        reqHeader.put("authorization", tokenType + " " + token);
        reqHeader.put("content-type", "application/x-www-form-urlencoded");

        //=========== ボディの生成 ===========
        String reqBody;
        try {
            //DTOからJSON文字列にマッピング
            ObjectMapper objectMapper = new ObjectMapper();
            reqBody = "request_data=" + objectMapper.writeValueAsString(srvSearchAreaReqDto);
        } catch (JsonProcessingException e) {
            Log.e(TAG,methodName + ":request body mapping error");
            throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== API実行 ===========
        ApiResult apiResult;
        try {
            apiResult = invokeApi(reqUrl, Method.POST, reqHeader, reqBody, HttpsURLConnection.HTTP_OK);
        } catch(SocketTimeoutException e) {
            //タイムアウト
            throw new AreaServerException(AreaServerException.ErrorType.TIMEOUT, e);
        } catch(Exception e) {
            //その他エラー
            throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
        }

        //=========== レスポンスの確認 ===========
        SrvSearchAreaResDto srvSearchAreaResDto;

        if (apiResult.httpStatusCode == HttpsURLConnection.HTTP_OK) {
            //正常レスポンス

            if(apiResult.rawResBody == null) {
                //レスポンスボディが空の場合はエラー
                Log.e(TAG,methodName + ":response body is null");
                throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR);
            }

            try {
                //JSON文字列をDTOにマッピング
                ObjectMapper objectMapper = new ObjectMapper();
                srvSearchAreaResDto = objectMapper.readValue(apiResult.rawResBody, SrvSearchAreaResDto.class);
            } catch (JsonProcessingException e) {
                Log.e(TAG,methodName + ":response body mapping error");
                throw new AreaServerException(AreaServerException.ErrorType.ETC_ERROR, e);
            }

        } else if(apiResult.httpStatusCode == HttpsURLConnection.HTTP_BAD_REQUEST) {
            //サーバーエラー（パラメータエラー）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new AreaServerException(AreaServerException.ErrorType.SERVER_PARAM_ERROR);

        } else {
            //サーバーエラー（上記以外）
            Log.e(TAG,methodName + ":response error / HTTP STATUS=" + apiResult.httpStatusCode);
            throw new AreaServerException(AreaServerException.ErrorType.SERVER_ERROR);
        }

        return srvSearchAreaResDto;
    }

    /**
     * API実行
     *
     * @param reqUrl 接続先URL
     * @param method HTTPメソッド
     * @param reqHeader リクエストヘッダ
     * @param reqBody リクエストボディ
     * @param successHttpStatusCode 成功とみなすHTTPステータスコード
     * @return API実行結果
     */
    private ApiResult invokeApi(final URL reqUrl,
                                @SuppressWarnings("SameParameterValue") final Method method,
                                final Map<String, String> reqHeader,
                                final String reqBody,
                                @SuppressWarnings("SameParameterValue") final int successHttpStatusCode) throws IOException {

        final String methodName = "invokeApi";

        Log.i(TAG, "RequestUrl:" +  reqUrl);
        Log.i(TAG, "RequestBody:" +  reqBody);

        HttpsURLConnection connection = null; // HTTPSコネクション
        OutputStream outputStream = null; // コネクションへの出力用Stream
        BufferedReader bufferedReader = null; // コネクションからボディデータを読み込む際に使用
        ApiResult result = new ApiResult();
        try {
            // 接続オブジェクトを生成
            connection = (HttpsURLConnection) reqUrl.openConnection();

            // セットアップパラメータを指定
            connection.setUseCaches(false); // 接続時にキャッシュを利用しない
            connection.setRequestMethod(method.name()); // HTTPメソッド
            connection.setConnectTimeout(CONNECTION_TIMEOUT); // コネクションタイムアウト
            connection.setReadTimeout(READ_TIMEOUT); // リードタイムアウト
            if (reqHeader != null) {
                for (String key : reqHeader.keySet()) {
                    Log.i(TAG, "RequestHeader:" +  key + "=" + reqHeader.get(key));
                    connection.setRequestProperty(key, reqHeader.get(key)); // リクエストヘッダ
                }
            }
            if (reqBody != null && !reqBody.isEmpty()) {
                connection.setDoOutput(true); // リクエストボディの送信を許可（コネクション確立前に設定しておく必要がある）
            }

            // コネクション確立
            connection.connect();

            if (reqBody != null && !reqBody.isEmpty()) {
                outputStream = connection.getOutputStream();
                outputStream.write(reqBody.getBytes(StandardCharsets.UTF_8));
                outputStream.close();
            }

            // レスポンスHTTPステータスコード
            // 補足：レスポンスが正常か異常かで、Streamの取得方法が異なる
            result.httpStatusCode = connection.getResponseCode();
            try (InputStream inputStream = successHttpStatusCode == result.httpStatusCode ?
                    connection.getInputStream() : connection.getErrorStream()
            ) {

                // レスポンスヘッダ
                for (String key : connection.getHeaderFields().keySet()) {
                    result.resHeader.put(key, connection.getHeaderField(key));
                }

                // レスポンスボディ
                StringBuilder response = new StringBuilder();
                bufferedReader = new BufferedReader(new InputStreamReader(inputStream, StandardCharsets.UTF_8));
                bufferedReader.lines().forEach(response::append);
                if (response.length() > 0) {
                    result.rawResBody = response.toString();
                }
            }

        } catch (SocketTimeoutException e) {
            // タイムアウト
            Log.e(TAG, methodName + ":SocketTimeoutException");
            throw e;
        } catch (IOException e) {
            e.printStackTrace();
            Log.e(TAG, methodName + ":IOException");
            throw e;
        } catch (Exception e) {
            e.printStackTrace();
            Log.e(TAG, methodName + ":Exception");
            throw e;
        } finally {
            try {
                if (outputStream != null) {
                    outputStream.close();
                }
                if (bufferedReader != null) {
                    bufferedReader.close();
                }
                if (connection != null) {
                    // コネクション切断
                    connection.disconnect();
                }
            } catch (IOException e) {
                // 本エラーが発生することはない
                e.printStackTrace();
                Log.e(TAG, methodName + ":IOException");
            }
        }

        Log.i(TAG, "HTTP Status:" +  result.httpStatusCode);
        Log.i(TAG, "ResponseBody:" +  result.rawResBody);

        return result;
    }

    /**
     * Base64エンコード
     *
     * @param src エンコード対象文字列
     * @return Base64エンコードされた文字列
     */
    private static String getBase64encode(String src) {
        return new String(Base64.encode(src.getBytes(), Base64.NO_WRAP));
    }
}
