package net.zenrindatacom.geofenceapp_sample.locationlib;

import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.text.TextUtils;

import net.zenrindatacom.geofenceapp_sample.R;
import net.zenrindatacom.geofenceapp_sample.application.SampleApplication;
import net.zenrindatacom.geofenceapp_sample.preference.LocationLibPreference;

import java.io.Serializable;
import java.util.ArrayList;

/**
 * 位置測位ライブラリの呼び出し
 */
public class LocationServiceManager implements Serializable {

    /**
     * 位置測位ライブラリ開始
     */
    public static void startLocationService() {
        Intent intent = new Intent();

        LocationLibPreference llPreference = LocationLibPreference.getInstance();

        //呼び出し先のコンポーネント名を設定
        setIntentLocationLibComponent(intent);

        //Intentの種類（Operation）：測位開始
        intent.putExtra(LocationLibConst.KEY_OPERATION, LocationLibConst.OPERATION_START);

        //プロバイダー：GPS
        boolean gps = llPreference.getProviderGps();
        //プロバイダー：ネットワーク
        boolean network = llPreference.getProviderNetwork();
        //プロバイダー種別（Provider）を配列型で格納する
        String[] provider = getProviderSetting(gps, network);
        if (provider != null) {
            intent.putExtra(LocationLibConst.KEY_PROVIDER, provider);
        }

        //測位間隔（Interval）
        String value = llPreference.getInterval();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_INTERVAL, Integer.valueOf(value));
        }

        //測位精度（AccuracyThreshold）
        value = llPreference.getAccuracyThreshold();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_ACCURACY_THRESHOLD, Float.valueOf(value));
        }

        //測位結果通知アプリのパッケージ名（PackageName and ClassName）
        setIntentNotifyPackageAndClassName(intent);

        //デバッグモード（Debug）
        boolean bValue =  llPreference.getDebug();
        intent.putExtra(LocationLibConst.KEY_DEBUG, bValue);

        //同一位置判定距離（SamePositionDeterminingDistance）
        value = llPreference.getSamePositionDeterminingDistance();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_SAME_POSITION_DETERMINING_DISTANCE, Integer.valueOf(value));
        }

        //同一位置判定回数（SamePositionDeterminingCountThreshold）
        value = llPreference.getSamePositionDeterminingCountThreshold();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_SAME_POSITION_DETERMINING_COUNT, Integer.valueOf(value));
        }

        //NotificationID
        intent.putExtra(LocationLibConst.KEY_NOTIFICATION_ID, 1);

        //Notification
        intent.putExtra(LocationLibConst.KEY_NOTIFICATION, createNotification());

        //APIキー（ApiKey）
        value = llPreference.getApiKey();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_API_KEY, value);
        }

        //クライアントID（ClientID）
        value = llPreference.getClientId();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_CLIENT_ID, value);
        }

        //秘密鍵（Secret）
        value = llPreference.getSecret();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_SECRET, value);
        }

        value = llPreference.getAuthEnv();
        if (!TextUtils.isEmpty(value)) {
            intent.putExtra(LocationLibConst.KEY_AUTH_ENV, Integer.valueOf(value));
        }

        //Android8以降はstartForegroundServiceでサービス起動
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            SampleApplication.getAppContext().startForegroundService(intent);
        } else {
            SampleApplication.getAppContext().startService(intent);
        }
    }

    /**
     * 位置測位ライブラリ停止
     */
    public static void stopLocationService() {
        Intent intent = new Intent();

        //呼び出し先のコンポーネント名を設定
        setIntentLocationLibComponent(intent);

        //Intentの種類（Operation）: 測位停止
        intent.putExtra(LocationLibConst.KEY_OPERATION, LocationLibConst.OPERATION_STOP);

        SampleApplication.getAppContext().startService(intent);
    }

    /**
     * 引数のIntentに呼び出し先Component名として、「位置測位ライブラリ」サービスを設定する
     *
     * @param intent 設定追加先のIntent
     */
    private static void setIntentLocationLibComponent(Intent intent) {
        String packageName = SampleApplication.getAppContext().getPackageName();
        intent.setComponent(new ComponentName(packageName,
                "net.zenrindatacom.location.request.service.LocationRequestService"));
    }

    /**
     * 引数のIntentに「位置測位ライブラリからの通知を受けるパッケージ名とクラス名」を設定する
     * 本サンプルではLocationLibReceiveServiceで受け取る
     *
     * @param intent パラメータ設定先のIntent
     */
    private static void setIntentNotifyPackageAndClassName(Intent intent) {
        Context context = SampleApplication.getAppContext();
        //パッケージ名（PackageName)
        intent.putExtra(LocationLibConst.KEY_PACKAGE_NAME, context.getPackageName());

        //クラス名（ClassName）
        intent.putExtra(LocationLibConst.KEY_CLASS_NAME, LocationLibReceiveService.class.getName());
    }

    /**
     * 引数「プロバイダー」設定を配列に変換して返す
     *
     * @param isGPS プロバイダー（GPS）のON、OFF
     * @param isNetwork プロバイダー（NETWORK）のON、OFF
     * @return Providerの文字列配列
     */
    private static String[] getProviderSetting(boolean isGPS, boolean isNetwork) {

        ArrayList<String> provider = new ArrayList<>();
        if (isGPS) {
            provider.add("GPS");
        }
        if (isNetwork) {
            provider.add("NETWORK");
        }

        return provider.size()==0 ? null : provider.toArray(new String[0]);
    }

    /**
     * 測位開始後Android端末に表示するNotificationを作成する
     *
     * @return 作成したNotification
     */
    private static Notification createNotification() {
        Context ctx = SampleApplication.getAppContext();
        Notification notification ;
        //タイトル
        String title = ctx.getString(R.string.service_message_title);
        //本文
        String text = ctx.getString(R.string.service_message_body);
        //アイコン
        int icon = R.mipmap.ic_launcher;

        NotificationManager manager = (NotificationManager) ctx.getSystemService(Context.NOTIFICATION_SERVICE);
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            //Android8以降はNotification channelの設定が必要
            NotificationChannel channel = new NotificationChannel(
                    // アプリでユニークな ID
                    "app_channel",
                    // ユーザが「設定」アプリで見ることになるチャンネル名
                    ctx.getString(R.string.service_channel_name),
                    // チャンネルの重要度（0 ~ 4）
                    NotificationManager.IMPORTANCE_LOW
            );
            manager.createNotificationChannel(channel);

            notification = new Notification.Builder(ctx, "app_channel")
                    .setContentTitle(title)
                    .setContentText(text)
                    .setSmallIcon(icon)
                    .setAutoCancel(true)
                    .build();
        } else {
            //Android8未満
            notification = new Notification.Builder(ctx)
                    .setContentTitle(title)
                    .setContentText(text)
                    .setSmallIcon(icon)
                    .setAutoCancel(true)
                    .build();
        }

        return notification;
    }
}
